File size: 12,761 Bytes
d1be154 | 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 32 33 34 35 36 37 38 39 40 41 42 43 44 45 46 47 48 49 50 51 52 53 54 55 56 57 58 59 60 61 62 63 64 65 66 67 68 69 70 71 72 73 74 75 76 77 78 79 80 81 82 83 84 85 86 87 88 89 90 91 92 93 94 95 96 97 98 99 100 101 102 103 104 105 106 107 108 109 110 111 112 113 114 115 116 117 118 119 120 121 122 123 124 125 126 127 128 129 130 131 132 133 134 135 136 137 138 139 140 141 142 143 144 145 146 147 148 149 150 151 152 153 154 155 156 157 158 159 160 161 162 163 164 165 166 167 168 169 170 171 172 173 174 175 176 177 178 179 180 181 182 183 184 185 186 187 188 189 190 191 192 193 194 195 196 197 198 199 200 201 202 203 204 205 206 207 208 209 210 211 212 213 214 215 216 217 218 219 220 221 222 223 224 225 226 227 228 229 230 231 232 233 234 235 236 237 238 239 240 241 242 243 244 245 246 247 248 249 250 251 252 253 254 255 256 257 258 259 260 261 262 263 264 265 266 267 268 269 270 271 272 273 274 275 276 277 278 279 280 281 282 283 284 285 286 287 288 289 290 291 292 293 294 295 296 297 298 299 300 301 302 303 304 305 306 307 308 309 310 311 312 313 314 315 316 317 318 319 320 321 322 323 324 325 326 327 328 329 330 331 332 333 334 335 336 337 338 339 340 341 342 343 344 345 346 347 348 349 350 351 352 353 354 355 356 357 358 359 360 361 362 363 364 365 366 367 368 369 370 371 372 373 374 375 376 377 378 379 380 381 382 383 384 385 386 387 388 389 390 391 392 393 394 395 396 397 398 399 400 401 402 403 404 405 406 407 408 409 410 411 412 413 414 415 416 417 418 419 420 421 422 423 424 425 426 427 428 429 430 431 432 433 434 435 436 437 438 439 440 441 442 443 444 445 446 447 448 449 450 451 452 453 454 455 456 457 458 459 460 461 462 463 464 465 466 467 468 469 470 471 472 473 474 475 476 477 478 479 480 481 482 483 484 485 486 487 488 489 | #pragma once
#include "reduce.hpp"
/**
@file taskflow/cuda/algorithm/scan.hpp
@brief CUDA scan algorithm include file
*/
namespace tf::detail {
// ----------------------------------------------------------------------------
// scan
// ----------------------------------------------------------------------------
/** @private */
inline constexpr unsigned cudaScanRecursionThreshold = 8;
/** @private */
enum class cudaScanType : int {
EXCLUSIVE = 1,
INCLUSIVE
};
/** @private */
template<typename T, unsigned vt = 0, bool is_array = (vt > 0)>
struct cudaScanResult {
T scan;
T reduction;
};
/** @private */
template<typename T, unsigned vt>
struct cudaScanResult<T, vt, true> {
cudaArray<T, vt> scan;
T reduction;
};
//-----------------------------------------------------------------------------
/** @private */
template<unsigned nt, typename T>
struct cudaBlockScan {
static constexpr unsigned num_warps = nt / CUDA_WARP_SIZE;
static constexpr unsigned num_passes = static_floor_log2<nt>();
static constexpr unsigned capacity = nt + num_warps;
/** @private */
union storage_t {
T data[2 * nt];
struct { T threads[nt], warps[num_warps]; };
};
// standard scan
template<typename op_t>
__device__ cudaScanResult<T> operator ()(
unsigned tid,
T x,
storage_t& storage,
unsigned count = nt,
op_t op = op_t(),
T init = T(),
cudaScanType type = cudaScanType::EXCLUSIVE
) const;
// vectorized scan. accepts multiple values per thread and adds in
// optional global carry-in
template<unsigned vt, typename op_t>
__device__ cudaScanResult<T, vt> operator()(
unsigned tid,
cudaArray<T, vt> x,
storage_t& storage,
T carry_in = T(),
bool use_carry_in = false,
unsigned count = nt,
op_t op = op_t(),
T init = T(),
cudaScanType type = cudaScanType::EXCLUSIVE
) const;
};
// standard scan
template <unsigned nt, typename T>
template<typename op_t>
__device__ cudaScanResult<T> cudaBlockScan<nt, T>::operator () (
unsigned tid, T x, storage_t& storage, unsigned count, op_t op,
T init, cudaScanType type
) const {
unsigned first = 0;
storage.data[first + tid] = x;
__syncthreads();
cuda_iterate<num_passes>([&](auto pass) {
if(auto offset = 1<<pass; tid >= offset) {
x = op(storage.data[first + tid - offset], x);
}
first = nt - first;
storage.data[first + tid] = x;
__syncthreads();
});
cudaScanResult<T> result;
result.reduction = storage.data[first + count - 1];
result.scan = (tid < count) ?
(cudaScanType::INCLUSIVE == type ? x :
(tid ? storage.data[first + tid - 1] : init)) :
result.reduction;
__syncthreads();
return result;
}
// vectorized scan block
template <unsigned nt, typename T>
template<unsigned vt, typename op_t>
__device__ cudaScanResult<T, vt> cudaBlockScan<nt, T>::operator()(
unsigned tid,
cudaArray<T, vt> x,
storage_t& storage,
T carry_in,
bool use_carry_in,
unsigned count, op_t op,
T init,
cudaScanType type
) const {
// Start with an inclusive scan of the in-range elements.
if(count >= nt * vt) {
cuda_iterate<vt>([&](auto i) {
x[i] = i ? op(x[i], x[i - 1]) : x[i];
});
} else {
cuda_iterate<vt>([&](auto i) {
auto index = vt * tid + i;
x[i] = i ?
((index < count) ? op(x[i], x[i - 1]) : x[i - 1]) :
(x[i] = (index < count) ? x[i] : init);
});
}
// Scan the thread-local reductions for a carry-in for each thread.
auto result = operator()(
tid, x[vt - 1], storage,
(count + vt - 1) / vt, op, init, cudaScanType::EXCLUSIVE
);
// Perform the scan downsweep and add both the global carry-in and the
// thread carry-in to the values.
if(use_carry_in) {
result.reduction = op(carry_in, result.reduction);
result.scan = tid ? op(carry_in, result.scan) : carry_in;
} else {
use_carry_in = tid > 0;
}
cudaArray<T, vt> y;
cuda_iterate<vt>([&](auto i) {
if(cudaScanType::EXCLUSIVE == type) {
y[i] = i ? x[i - 1] : result.scan;
if(use_carry_in && i > 0) y[i] = op(result.scan, y[i]);
} else {
y[i] = use_carry_in ? op(x[i], result.scan) : x[i];
}
});
return cudaScanResult<T, vt> { y, result.reduction };
}
/**
@private
@brief single-pass scan for small input
*/
template <typename P, typename I, typename O, typename C>
void cuda_single_pass_scan(
P&& p,
cudaScanType scan_type,
I input,
unsigned count,
O output,
C op
//reduction_it reduction,
) {
using T = typename std::iterator_traits<O>::value_type;
using E = std::decay_t<P>;
// Small input specialization. This is the non-recursive branch.
cuda_kernel<<<1, E::nt, 0, p.stream()>>>([=] __device__ (auto tid, auto bid) {
using scan_t = cudaBlockScan<E::nt, T>;
__shared__ union {
typename scan_t::storage_t scan;
T values[E::nv];
} shared;
auto carry_in = T();
for(unsigned cur = 0; cur < count; cur += E::nv) {
// Cooperatively load values into register.
auto count2 = min(count - cur, E::nv);
auto x = cuda_mem_to_reg_thread<E::nt, E::vt>(input + cur,
tid, count2, shared.values);
auto result = scan_t()(tid, x, shared.scan,
carry_in, cur > 0, count2, op, T(), scan_type);
// Store the scanned values back to global memory.
cuda_reg_to_mem_thread<E::nt, E::vt>(result.scan, tid, count2,
output + cur, shared.values);
// Roll the reduction into carry_in.
carry_in = result.reduction;
}
// Store the carry-out to the reduction pointer. This may be a
// discard_iterator_t if no reduction is wanted.
//if(!tid) *reduction = carry_in;
});
}
/**
@private
@brief main scan loop
*/
template<typename P, typename I, typename O, typename C>
void cuda_scan_loop(
P&& p,
cudaScanType scan_type,
I input,
unsigned count,
O output,
C op,
//reduction_it reduction,
void* ptr
) {
using E = std::decay_t<P>;
using T = typename std::iterator_traits<O>::value_type;
T* buffer = static_cast<T*>(ptr);
//launch_t::cta_dim(context).B(count);
unsigned B = (count + E::nv - 1) / E::nv;
if(B > cudaScanRecursionThreshold) {
//cudaDeviceVector<T> partials(B);
//auto buffer = partials.data();
// upsweep phase
cuda_kernel<<<B, E::nt, 0, p.stream()>>>([=] __device__ (auto tid, auto bid) {
__shared__ typename cudaBlockReduce<E::nt, T>::Storage shm;
// Load the tile's data into register.
auto tile = cuda_get_tile(bid, E::nv, count);
auto x = cuda_mem_to_reg_strided<E::nt, E::vt>(
input + tile.begin, tid, tile.count()
);
// Reduce the thread's values into a scalar.
T scalar;
cuda_strided_iterate<E::nt, E::vt>(
[&] (auto i, auto j) { scalar = i ? op(scalar, x[i]) : x[0]; },
tid, tile.count()
);
// Reduce across all threads.
auto all_reduce = cudaBlockReduce<E::nt, T>()(
tid, scalar, shm, tile.count(), op
);
// Store the final reduction to the partials.
if(!tid) {
buffer[bid] = all_reduce;
}
});
// recursively call scan
//cuda_scan_loop(p, cudaScanType::EXCLUSIVE, buffer, B, buffer, op, S);
cuda_scan_loop(
p, cudaScanType::EXCLUSIVE, buffer, B, buffer, op, buffer+B
);
// downsweep: perform an intra-tile scan and add the scan of the partials
// as carry-in
cuda_kernel<<<B, E::nt, 0, p.stream()>>>([=] __device__ (auto tid, auto bid) {
using scan_t = cudaBlockScan<E::nt, T>;
__shared__ union {
typename scan_t::storage_t scan;
T values[E::nv];
} shared;
// Load a tile to register in thread order.
auto tile = cuda_get_tile(bid, E::nv, count);
auto x = cuda_mem_to_reg_thread<E::nt, E::vt>(
input + tile.begin, tid, tile.count(), shared.values
);
// Scan the array with carry-in from the partials.
auto y = scan_t()(tid, x, shared.scan,
buffer[bid], bid > 0, tile.count(), op, T(),
scan_type).scan;
// Store the scanned values to the output.
cuda_reg_to_mem_thread<E::nt, E::vt>(
y, tid, tile.count(), output + tile.begin, shared.values
);
});
}
// Small input specialization. This is the non-recursive branch.
else {
cuda_single_pass_scan(p, scan_type, input, count, output, op);
}
}
} // namespace tf::detail ----------------------------------------------------
namespace tf {
// Function: scan_bufsz
template <unsigned NT, unsigned VT>
template <typename T>
unsigned cudaExecutionPolicy<NT, VT>::scan_bufsz(unsigned count) {
unsigned B = num_blocks(count);
unsigned n = 0;
for(auto b=B; b>detail::cudaScanRecursionThreshold; b=num_blocks(b)) {
n += b;
}
return n*sizeof(T);
}
/**
@brief performs asynchronous inclusive scan over a range of items
@tparam P execution policy type
@tparam I input iterator
@tparam O output iterator
@tparam C binary operator type
@param p execution policy
@param first iterator to the beginning of the input range
@param last iterator to the end of the input range
@param output iterator to the beginning of the output range
@param op binary operator to apply to scan
@param buf pointer to the temporary buffer
*/
template<typename P, typename I, typename O, typename C>
void cuda_inclusive_scan(
P&& p, I first, I last, O output, C op, void* buf
) {
unsigned count = std::distance(first, last);
if(count == 0) {
return;
}
// launch the scan loop
detail::cuda_scan_loop(
p, detail::cudaScanType::INCLUSIVE, first, count, output, op, buf
);
}
/**
@brief performs asynchronous inclusive scan over a range of transformed items
@tparam P execution policy type
@tparam I input iterator
@tparam O output iterator
@tparam C binary operator type
@tparam U unary operator type
@param p execution policy
@param first iterator to the beginning of the input range
@param last iterator to the end of the input range
@param output iterator to the beginning of the output range
@param bop binary operator to apply to scan
@param uop unary operator to apply to transform each item before scan
@param buf pointer to the temporary buffer
*/
template<typename P, typename I, typename O, typename C, typename U>
void cuda_transform_inclusive_scan(
P&& p, I first, I last, O output, C bop, U uop, void* buf
) {
using T = typename std::iterator_traits<O>::value_type;
unsigned count = std::distance(first, last);
if(count == 0) {
return;
}
// launch the scan loop
detail::cuda_scan_loop(
p, detail::cudaScanType::INCLUSIVE,
cuda_make_load_iterator<T>([=]__device__(auto i){ return uop(*(first+i)); }),
count, output, bop, buf
);
}
/**
@brief performs asynchronous exclusive scan over a range of items
@tparam P execution policy type
@tparam I input iterator
@tparam O output iterator
@tparam C binary operator type
@param p execution policy
@param first iterator to the beginning of the input range
@param last iterator to the end of the input range
@param output iterator to the beginning of the output range
@param op binary operator to apply to scan
@param buf pointer to the temporary buffer
*/
template<typename P, typename I, typename O, typename C>
void cuda_exclusive_scan(
P&& p, I first, I last, O output, C op, void* buf
) {
unsigned count = std::distance(first, last);
if(count == 0) {
return;
}
// launch the scan loop
detail::cuda_scan_loop(
p, detail::cudaScanType::EXCLUSIVE, first, count, output, op, buf
);
}
/**
@brief performs asynchronous exclusive scan over a range of items
@tparam P execution policy type
@tparam I input iterator
@tparam O output iterator
@tparam C binary operator type
@tparam U unary operator type
@param p execution policy
@param first iterator to the beginning of the input range
@param last iterator to the end of the input range
@param output iterator to the beginning of the output range
@param bop binary operator to apply to scan
@param uop unary operator to apply to transform each item before scan
@param buf pointer to the temporary buffer
*/
template<typename P, typename I, typename O, typename C, typename U>
void cuda_transform_exclusive_scan(
P&& p, I first, I last, O output, C bop, U uop, void* buf
) {
using T = typename std::iterator_traits<O>::value_type;
unsigned count = std::distance(first, last);
if(count == 0) {
return;
}
// launch the scan loop
detail::cuda_scan_loop(
p, detail::cudaScanType::EXCLUSIVE,
cuda_make_load_iterator<T>([=]__device__(auto i){ return uop(*(first+i)); }),
count, output, bop, buf
);
}
} // end of namespace tf -----------------------------------------------------
|